from dataclasses import dataclass
from typing import Any
import re
from collections import defaultdict
from datetime import datetime, date, time


@dataclass(repr=False)
class FDBType:
    """
    Holds information about how to format and validate a given FDB Schema type like Time or Expver
    This base type represents a string and does no validation or formatting. It's the default type.
    """

    name: str = "String"

    def __repr__(self) -> str:
        return self.name

    def validate(self, s: Any) -> bool:
        try:
            self.parse(s)
            return True
        except (ValueError, AssertionError):
            return False

    def format(self, s: Any) -> str:
        return str(s).lower()

    def parse(self, s: str) -> Any:
        return s


@dataclass(repr=False)
class Expver_FDBType(FDBType):
    name: str = "Expver"

    def parse(self, s: str) -> str:
        assert bool(re.match(".{4}", s))
        return s


@dataclass(repr=False)
class Time_FDBType(FDBType):
    name: str = "Time"
    time_format = "%H%M"

    def format(self, t: time) -> str:
        return t.strftime(self.time_format)

    def parse(self, s: datetime | str | int) -> time:
        if isinstance(s, str):
            assert len(s) == 4
            return datetime.strptime(s, self.time_format).time()
        if isinstance(s, datetime):
            return s.time()
        return self.parse(f"{s:04}")


@dataclass(repr=False)
class Date_FDBType(FDBType):
    name: str = "Date"
    date_format: str = "%Y%m%d"

    def format(self, d: Any) -> str:
        if isinstance(d, date):
            return d.strftime(self.date_format)
        if isinstance(d, int):
            return f"{d:08}"
        else:
            return d

    def parse(self, s: datetime | str | int) -> date:
        if isinstance(s, str):
            return datetime.strptime(s, self.date_format).date()
        elif isinstance(s, datetime):
            return s.date()
        return self.parse(f"{s:08}")


FDB_type_to_implementation = defaultdict(lambda: FDBType()) | {
    cls.name: cls() for cls in [Expver_FDBType, Time_FDBType, Date_FDBType]
}



import dataclasses
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from typing import Any, Callable, Iterable, Literal


@dataclass(frozen=True)
class HTML():
    html: str
    def _repr_html_(self):
        return self.html

@dataclass(frozen=True)
class Values(ABC):
    @abstractmethod
    def summary(self) -> str:
        pass
    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __contains__(self, value: Any) -> bool:
        pass

    @abstractmethod
    def from_strings(self, values: list[str]) -> list['Values']:
        pass

@dataclass(frozen=True)
class Enum(Values):
    """
    The simplest kind of key value is just a list of strings.
    summary -> string1/string2/string....
    """
    values: list[Any]

    def __len__(self) -> int:
        return len(self.values)
    def summary(self) -> str:
        return '/'.join(sorted(self.values))
    def __contains__(self, value: Any) -> bool:
        return value in self.values
    def from_strings(self, values: list[str]) -> list['Values']:
        return [Enum(values)]

@dataclass(frozen=True)
class Range(Values, ABC):
    dtype: str = dataclasses.field(kw_only=True)

@dataclass(frozen=True)
class DateRange(Range):
    start: date
    end: date
    step: timedelta
    dtype: Literal["date"] = dataclasses.field(kw_only=True, default="date")

    @classmethod
    def from_strings(self, values: list[str]) -> list['DateRange']:
        dates = sorted([datetime.strptime(v, "%Y%m%d") for v in values])
        if len(dates) < 2:
            return [DateRange(
                start=dates[0],
                end=dates[0],
                step=timedelta(days=0)
            )]
        
        ranges = []
        current_range, dates = [dates[0],], dates[1:]
        while len(dates) > 1:
            if dates[0] - current_range[-1] == timedelta(days=1):
                current_range.append(dates.pop(0))
            
            elif len(current_range) == 1:
                ranges.append(DateRange(
                start=current_range[0],
                end=current_range[0],
                step=timedelta(days=0)
                ))
                current_range = [dates.pop(0),]

            else:
                ranges.append(DateRange(
                start=current_range[0],
                end=current_range[-1],
                step=timedelta(days=1)
                ))
                current_range = [dates.pop(0),]
        return ranges
    
    def __contains__(self, value: Any) -> bool:
        v = datetime.strptime(value, "%Y%m%d").date()
        return self.start <= v <= self.end and (v - self.start) % self.step == 0


    def __len__(self) -> int:
        return (self.end - self.start) // self.step
    
    def summary(self) -> str:
        def fmt(d): return d.strftime("%Y%m%d")
        if self.step == timedelta(days=0):
            return f"{fmt(self.start)}"
        if self.step == timedelta(days=1):
            return f"{fmt(self.start)}/to/{fmt(self.end)}"
        
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step // timedelta(days=1)}"

@dataclass(frozen=True)
class TimeRange(Range):
    start: int
    end: int
    step: int
    dtype: Literal["time"] = dataclasses.field(kw_only=True, default="time")

    @classmethod
    def from_strings(self, values: list[str]) -> list['TimeRange']:
        if len(values) == 0: return []

        times = sorted([int(v) for v in values])
        if len(times) < 2:
            return [TimeRange(
                start=times[0],
                end=times[0],
                step=100
            )]
        
        ranges = []
        current_range, times = [times[0],], times[1:]
        while len(times) > 1:
            if times[0] - current_range[-1] == 1:
                current_range.append(times.pop(0))
            
            elif len(current_range) == 1:
                ranges.append(TimeRange(
                start=current_range[0],
                end=current_range[0],
                step=0
                ))
                current_range = [times.pop(0),]

            else:
                ranges.append(TimeRange(
                start=current_range[0],
                end=current_range[-1],
                step=1
                ))
                current_range = [times.pop(0),]
        return ranges

    def __len__(self) -> int:
        return (self.end - self.start) // self.step
    
    def summary(self) -> str:
        def fmt(d): return f"{d:04d}"
        if self.step == 0:
            return f"{fmt(self.start)}"
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step}"
    
    def __contains__(self, value: Any) -> bool:
        v = int(value)
        return self.start <= v <= self.end and (v - self.start) % self.step == 0

@dataclass(frozen=True)
class IntRange(Range):
    dtype: Literal["int"]
    start: int
    end: int
    step: int
    dtype: Literal["int"] = dataclasses.field(kw_only=True, default="int")

    def __len__(self) -> int:
        return (self.end - self.start) // self.step
    
    def summary(self) -> str:
        def fmt(d): return d.strftime("%Y%m%d")
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step}"
    
    def __contains__(self, value: Any) -> bool:
        v = int(value)
        return self.start <= v <= self.end and (v - self.start) % self.step == 0
    

def values_from_json(obj) -> Values:
    if isinstance(obj, list): 
        return Enum(obj)

    match obj["dtype"]:
        case "date": return DateRange(**obj)
        case "time": return TimeRange(**obj)
        case "int": return IntRange(**obj)
        case _: raise ValueError(f"Unknown dtype {obj['dtype']}")

@dataclass(frozen=True)
class Node:
    key: str
    values: Values # Must support len()
    metadata: dict[str, str] # Applies to all children
    payload: list[Any] # List of size product(len(n.values) for n in  ancestors(self))
    children: list['Node']

def summarize_node(node: Node) -> tuple[str, Node]:
    """
    Extracts a summarized representation of the node while collapsing single-child paths.
    Returns the summary string and the last node in the chain that has multiple children.
    """
    summary = []
    
    while True:
        values_summary = node.values.summary()
        if len(values_summary) > 50:
            values_summary = values_summary[:50] + "..."
        summary.append(f"{node.key}={values_summary}")

        # Move down if there's exactly one child, otherwise stop
        if len(node.children) != 1:
            break
        node = node.children[0]

    return ", ".join(summary), node

def node_tree_to_string(node : Node, prefix : str = "", depth = None) -> Iterable[str]:
    summary, node = summarize_node(node)
    
    if depth is not None and depth <= 0:
        yield summary + " - ...\n"
        return
    # Special case for nodes with only a single child, this makes the printed representation more compact
    elif len(node.children) == 1:
        yield summary + ", "
        yield from node_tree_to_string(node.children[0], prefix, depth = depth)
        return
    else:
        yield summary + "\n"
    
    for index, child in enumerate(node.children):
        connector = "└── " if index == len(node.children) - 1 else "├── "
        yield prefix + connector
        extension = "    " if index == len(node.children) - 1 else "│   "
        yield from node_tree_to_string(child, prefix + extension, depth = depth - 1 if depth is not None else None)

def node_tree_to_html(node : Node, prefix : str = "", depth = 1, connector = "") -> Iterable[str]:
    summary, node = summarize_node(node)
    
    if len(node.children) == 0:
        yield f'<span class="leaf">{connector}{summary}</span>'
        return
    else:
        open = "open" if depth > 0 else ""
        yield f"<details {open}><summary>{connector}{summary}</summary>"

    for index, child in enumerate(node.children):
        connector = "└── " if index == len(node.children) - 1 else "├── "
        extension = "    " if index == len(node.children) - 1 else "│   "
        yield from node_tree_to_html(child, prefix + extension, depth = depth - 1, connector = prefix+connector)
    yield "</details>"

@dataclass(frozen=True)
class CompressedTree:
    root: Node

    @classmethod
    def from_json(cls, json: dict) -> 'CompressedTree':
        def from_json(json: dict) -> Node:
            return Node(
                key=json["key"],
                values=values_from_json(json["values"]),
                metadata=json["metadata"] if "metadata" in json else {},
                payload=json["payload"] if "payload" in json else [],
                children=[from_json(c) for c in json["children"]]
            )
        return CompressedTree(root=from_json(json))
    
    def __str__(self):
        return "".join(node_tree_to_string(node=self.root))
    
    def html(self, depth = 2) -> HTML:
        return HTML(self._repr_html_(depth = depth))
    
    def _repr_html_(self, depth = 2):
        css = """
        <style>
        .qubed-tree-view {
            font-family: monospace;
            white-space: pre;
        }
        .qubed-tree-view details {
            # display: inline;
            margin-left: 0;
        }
        .qubed-tree-view summary {
            list-style: none;
            cursor: pointer;
            text-overflow: ellipsis;
            overflow: hidden;
            text-wrap: nowrap;
            display: block;
        }

        .qubed-tree-view .leaf {
            text-overflow: ellipsis;
            overflow: hidden;
            text-wrap: nowrap;
            display: block;
        }

        .qubed-tree-view summary:hover,span.leaf:hover {
            background-color: #f0f0f0;
        }
        .qubed-tree-view details > summary::after {
            content: ' ';
        }
        .qubed-tree-view details:not([open]) > summary::after {
            content: " ▼";
        }
        </style>

        """
        nodes = "".join(cc for c in self.root.children for cc in node_tree_to_html(node=c, depth=depth))
        return f"{css}<pre class='qubed-tree-view'>{nodes}</pre>"
    
    def print(self, depth = None):
        print("".join(cc for c in self.root.children for cc in node_tree_to_string(node=c, depth = depth)))

    def transform(self, func: Callable[[Node], Node]) -> 'CompressedTree':
        "Call a function on every node of the tree, any changes to the children of a node will be ignored."
        def transform(node: Node) -> Node:
            new_node = func(node)
            return dataclasses.replace(new_node, children = [transform(c) for c in node.children])
        return CompressedTree(root=transform(self.root))

    def guess_datatypes(self) -> 'CompressedTree':
        def guess_datatypes(node: Node) -> list[Node]:
            # Try to convert enum values into more structured types
            children = [cc for c in node.children for cc in guess_datatypes(c)]

            if isinstance(node.values, Enum):
                match node.key:
                    case "time": range_class = TimeRange
                    case "date": range_class = DateRange
                    case _: range_class = None

                if range_class is not None:
                    return [
                        dataclasses.replace(node, values = range, children = children)
                        for range in range_class.from_strings(node.values.values)
                    ]
            return [dataclasses.replace(node, children = children)]

        children = [cc for c in self.root.children for cc in guess_datatypes(c)]
        return CompressedTree(root=dataclasses.replace(self.root, children = children))

    
    def select(self, selection : dict[str, str | list[str]], mode: Literal["strict", "relaxed"] = "relaxed") -> 'CompressedTree':
        # make all values lists
        selection = {k : v if isinstance(v, list) else [v] for k,v in selection.items()}

        def not_none(xs): return [x for x in xs if x is not None]

        def select(node: Node) -> Node | None: 
            # Check if the key is specified in the selection
            if node.key not in selection: 
                if mode == "strict":
                    return None
                return dataclasses.replace(node, children = not_none(select(c) for c in node.children))
            
            # If the key is specified, check if any of the values match
            values = Enum([ c for c in selection[node.key] if c in node.values])

            if not values: 
                return None 
            
            return dataclasses.replace(node, values = values, children = not_none(select(c) for c in node.children))
            
        return CompressedTree(root=dataclasses.replace(self.root, children = not_none(select(c) for c in self.root.children)))
    
    def to_list_of_cubes(self):
        def to_list_of_cubes(node: Node) -> list[list[Node]]:
            return [[node] + sub_cube for c in node.children for sub_cube in to_list_of_cubes(c)]

        return to_list_of_cubes(self.root)

    def info(self):
        cubes = self.to_list_of_cubes()
        print(f"Number of distinct paths: {len(cubes)}")

        

        
# What should the interace look like?

# tree = CompressedTree.from_json(...)
# tree = CompressedTree.from_protobuf(...)

# tree.print(depth = 5) # Prints a nice tree representation


import dataclasses
from collections import defaultdict
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any, Callable, Hashable, Literal, Mapping

from frozendict import frozendict

from .tree_formatters import HTML, node_tree_to_html, node_tree_to_string
from .value_types import DateRange, Enum, IntRange, TimeRange, Values


def values_from_json(obj) -> Values:
    if isinstance(obj, list): 
        return Enum(tuple(obj))

    match obj["dtype"]:
        case "date": return DateRange(**obj)
        case "time": return TimeRange(**obj)
        case "int": return IntRange(**obj)
        case _: raise ValueError(f"Unknown dtype {obj['dtype']}")

# In practice use a frozendict
Metadata = Mapping[str, str | int | float | bool]

@dataclass(frozen=True, eq=True, order=True)
class NodeData:
    key: str
    values: Values
    metadata: dict[str, tuple[Hashable, ...]] = field(default_factory=frozendict, compare=False)

    def summary(self) -> str:
        return f"{self.key}={self.values.summary()}" if self.key != "root" else "root"

@dataclass(frozen=True, eq=True, order=True)
class Qube:
    data: NodeData
    children: tuple['Qube', ...]

    @property
    def key(self) -> str:
        return self.data.key
    
    @property
    def values(self) -> Values:
        return self.data.values
    
    @property
    def metadata(self) -> frozendict[str, Any]:
        return self.data.metadata

    
    def summary(self) -> str:
        return self.data.summary()
    
    @classmethod
    def make(cls, key : str, values : Values, children, **kwargs) -> 'Qube':
        return cls(
            data = NodeData(key, values,  metadata = kwargs.get("metadata", frozendict())
            ),
            children = tuple(sorted(children)),
        )


    @classmethod
    def from_json(cls, json: dict) -> 'Qube':
        def from_json(json: dict) -> Qube:
            return Qube.make(
                key=json["key"],
                values=values_from_json(json["values"]),
                metadata=json["metadata"] if "metadata" in json else {},
                children=tuple(from_json(c) for c in json["children"])
            )
        return from_json(json)
    
    @classmethod
    def from_dict(cls, d: dict) -> 'Qube':
        def from_dict(d: dict) -> tuple[Qube, ...]:
            return tuple(Qube.make(
                key=k.split("=")[0],
                values=Enum(tuple(k.split("=")[1].split("/"))),
                children=from_dict(children)
            ) for k, children in d.items())
        
        return Qube.make(key = "root",
                              values=Enum(("root",)),
                              children = from_dict(d))
    
    @classmethod
    def empty(cls) -> 'Qube':
        return cls.make("root", Enum(("root",)), [])

    
    def __str__(self, depth = None) -> str:
        return "".join(node_tree_to_string(node=self, depth = depth))
    
    def print(self, depth = None): print(self.__str__(depth = depth))
    
    def html(self, depth = 2, collapse = True) -> HTML:
        return HTML(node_tree_to_html(self, depth = depth, collapse = collapse))
    
    def _repr_html_(self) -> str:
        return node_tree_to_html(self, depth = 2, collapse = True)

    
    def __getitem__(self, args) -> 'Qube':
        key, value = args
        for c in self.children:
            if c.key == key and value in c.values:
                data = dataclasses.replace(c.data, values = Enum((value,)))
                return dataclasses.replace(c, data = data)
        raise KeyError(f"Key {key} not found in children of {self.key}")

    

    def transform(self, func: 'Callable[[Qube], Qube | list[Qube]]') -> 'Qube':
        """
        Call a function on every node of the Qube, return one or more nodes.
        If multiple nodes are returned they each get a copy of the (transformed) children of the original node.
        Any changes to the children of a node will be ignored.
        """
        def transform(node: Qube) -> list[Qube]:
            children = [cc for c in node.children for cc in transform(c)]
            new_nodes = func(node)
            if isinstance(new_nodes, Qube):
                new_nodes = [new_nodes]

            return [dataclasses.replace(new_node, children = children)
                    for new_node in new_nodes]
        
        children = tuple(cc for c in self.children for cc in transform(c))
        return dataclasses.replace(self, children = children)

    
    def select(self, selection : dict[str, str | list[str]], mode: Literal["strict", "relaxed"] = "relaxed") -> 'Qube':
        # make all values lists
        selection = {k : v if isinstance(v, list) else [v] for k,v in selection.items()}

        def not_none(xs): return tuple(x for x in xs if x is not None)

        def select(node: Qube) -> Qube | None: 
            # Check if the key is specified in the selection
            if node.key not in selection: 
                if mode == "strict":
                    return None
                return dataclasses.replace(node, children = not_none(select(c) for c in node.children))
            
            # If the key is specified, check if any of the values match
            values = Enum(tuple(c for c in selection[node.key] if c in node.values))

            if not values: 
                return None 
            
            return dataclasses.replace(node, values = values, children = not_none(select(c) for c in node.children))
            
        return dataclasses.replace(self, children = not_none(select(c) for c in self.children))
    

    @staticmethod
    def _insert(position: "Qube", identifier : list[tuple[str, list[str]]]):
        """
        This algorithm goes as follows:
        We're at a particular node in the Qube, and we have a list of key-values pairs that we want to insert.
        We take the first key values pair
        key, values = identifier.pop(0)

        The general idea is to insert key, values into the current node and use recursion to handle the rest of the identifier.
        
        We have two sources of values with possible overlap. The values to insert and the values attached to the children of this node.
        For each value coming from either source we put it in one of three categories:
            1) Values that exist only in the already existing child. (Coming exclusively from position.children)
            2) Values that exist in both a child and the new values.
            3) Values that exist only in the new values.
            

        Thus we add the values to insert to a set, and loop over the children.
        For each child we partition its values into the three categories.

        For 1) we create a new child node with the key, reduced set of values and the same children.
        For 2)
            Create a new child node with the key, and the values in group 2
            Recurse to compute the children

        Once we have finished looping over children we know all the values left over came exclusively from the new values.
        So we:
            Create a new node with these values.
            Recurse to compute the children

        Finally we return the node with all these new children.
        """
        pass
        # if not identifier:
        #     return position

        # key, values = identifier.pop(0)
        # # print(f"Inserting {key}={values} into {position.summary()}")

        # # Only the children with the matching key are relevant.
        # source_children = {c : [] for c in position.children if c.key == key}
        # new_children = []

        # values = set(values)
        # for c in source_children:
        #     values_set = set(c.values)
        #     group_1 = values_set - values
        #     group_2 = values_set & values
        #     values = values - values_set # At the end of this loop values will contain only the new values

        #     if group_1:
        #         group_1_node = Qube.make(c.key, Enum(tuple(group_1)), c.children)
        #         new_children.append(group_1_node) # Add the unaffected part of this child
            
        #     if group_2:
        #         new_node = Qube.make(key, Enum(tuple(affected)), [])
        #         new_node = Qube._insert(new_node, identifier)
        #         new_children.append(new_node) # Add the affected part of this child


        #     unaffected = [x for x in c.values if x not in affected]


        #     if affected: # This check is not technically necessary, but it makes the code more readable


        # # If there are any values not in any of the existing children, add them as a new child
        # if entirely_new_values:
        #     new_node = Qube.make(key, Enum(tuple(entirely_new_values)), [])
        #     new_children.append(Qube._insert(new_node, identifier))

        return Qube.make(position.key, position.values, new_children)

    def insert(self, identifier : dict[str, list[str]]) -> 'Qube':
        insertion = [(k, v) for k, v in identifier.items()]
        return Qube._insert(self, insertion)
    
    def to_list_of_cubes(self):
        def to_list_of_cubes(node: Qube) -> list[list[Qube]]:
            return [[node] + sub_cube for c in node.children for sub_cube in to_list_of_cubes(c)]

        return to_list_of_cubes(self)

    def info(self):
        cubes = self.to_list_of_cubes()
        print(f"Number of distinct paths: {len(cubes)}")

    @cached_property
    def structural_hash(self) -> int:
        """
        This hash takes into account the key, values and children's key values recursively.
        Because nodes are immutable, we only need to compute this once.
        """
        def hash_node(node: Qube) -> int:
            return hash((node.key, node.values, tuple(c.structural_hash for c in node.children)))

        return hash_node(self)

    def compress(self) -> "Qube":
        # First compress the children
        new_children = [child.compress() for child in self.children]

        # Now take the set of new children and see if any have identical key, metadata and children
        # the values may different and will be collapsed into a single node
        identical_children = defaultdict(set)
        for child in new_children:
            # only care about the key and children of each node, ignore values
            key = hash((child.key, tuple((cc.structural_hash for cc in child.children))))
            identical_children[key].add(child)
        
        # Now go through and create new compressed nodes for any groups that need collapsing
        new_children = []
        for child_set in identical_children.values():
            if len(child_set) > 1:
                child_set = list(child_set)
                key = child_set[0].key

                # Compress the children into a single node
                assert all(isinstance(child.data.values, Enum) for child in child_set), "All children must have Enum values"
                
                node_data = NodeData(
                    key = key,
                    metadata = frozendict(), # Todo: Implement metadata compression
                    values = Enum(tuple(v for child in child_set for v in child.data.values.values)),
                )
                new_child = Qube(data = node_data, children = child_set[0].children)
            else:
                # If the group is size one just keep it
                new_child = child_set.pop()
            
            new_children.append(new_child)

        return Qube(
            data = self.data,
            children = tuple(sorted(new_children))
        )


import argparse

# A simple command line app that reads from standard input and writes to standard output
# Arguments:
#    --input_format=fdb/mars
#    --output_format=text/html
import sys


def main():
    parser = argparse.ArgumentParser(description="Generate a compressed tree from various inputs.")
    
    parser.add_argument(
        "--input_format",
        choices=["fdb", "mars"],
        default="fdb",
        help="Specify the input format (fdb list or mars)."
    )
    
    parser.add_argument(
        "--output_format",
        choices=["text", "html"],
        default="text",
        help="Specify the output format (text or html)."
    )
    
    args = parser.parse_args()
    
    # Read from standard input
    l = 0
    for line in sys.stdin.readlines():
        l += 1
        
    
    # Process data (For now, just echoing the input)
    output_data = f"[Input Format: {args.input_format}] [Output Format: {args.output_format}]\n{l} lines read from standard input\n"
    
    # Write to standard output
    sys.stdout.write(output_data)

if __name__ == "__main__":
    main()


import json
from collections import defaultdict
from dataclasses import asdict, dataclass
from pathlib import Path

Tree = dict[str, "Tree"]

class RefcountedDict(dict[str, int]):
    refcount: int = 1

    def __repr__(self):
        return f"RefcountedDict(refcount={self.refcount}, {super().__repr__()})"
    
    def __hash__(self):
        return hash(tuple(sorted(self.items())))

@dataclass
class JSONNode:
    key: str
    values: list[str]
    children: list["JSONNode"]

class CompressedTree():
    """
    A implementation of a compressed tree that supports lookup, insertion, deletion and caching.
    The caching means that identical subtrees are stored only once, saving memory
    This is implemented internal by storing all subtrees in a global hash table
    
    """
    cache: dict[int, RefcountedDict]
    tree: RefcountedDict

    def _add_to_cache(self, level : RefcountedDict) -> int:
        "Add a level {key -> hash} to the cache"
        h = hash(level)
        if h not in self.cache:
            # Increase refcounts of the child nodes
            for child_h in level.values():
                self.cache[child_h].refcount += 1
            self.cache[h] = RefcountedDict(level)
        else:
            self.cache[h].refcount += 1
        return h
    
    def _replace_in_cache(self, old_h, level : RefcountedDict) -> int:
        """
        Replace the object at old_h with a different object level
        If the objects this is a no-op
        """
        # Start by adding the new object to the cache
        new_h = self._add_to_cache(level)

        # Now check if the old object needs to be garbage collected
        self._decrease_refcount(old_h)

        return new_h
    
    def _decrease_refcount(self, h : int):
        self.cache[h].refcount -= 1
        if self.cache[h].refcount == 0:
            # Recursively decrease refcounts of child nodes
            for child_h in self.cache[h].values():
                self._decrease_refcount(child_h)
            del self.cache[h]

    def cache_tree(self, tree : Tree) -> int:
        "Insert the given tree  (dictonary of dictionaries) (all it's children, recursively) into the hash table and return the hash key"
        level = RefcountedDict({k : self.cache_tree(v) for k, v in tree.items()})
        return self._add_to_cache(level)

    
    def _cache_path(self, path : list[str]) -> int:
        "Treat path = [x, y, z...] like {x : {y : {z : ...}}} and cache that"
        if not path: 
            return self.empty_hash
        k, *rest = path
        return self._add_to_cache(RefcountedDict({k : self._cache_path(rest)}))
    
    def reconstruct(self, max_depth = None) -> dict[str, dict]:
        "Reconstruct the tree as a normal nested dictionary"
        def reconstruct_node(h : int, depth : int) -> dict[str, dict]:
            if max_depth is not None and depth > max_depth:
                return {}
            return {k : reconstruct_node(v, depth+1) for k, v in self.cache[h].items()}
        return reconstruct_node(self.root_hash, 0)
    
    def reconstruct_compressed(self) -> dict[str, dict]:
        "Reconstruct the tree as a normal nested dictionary"
        def reconstruct_node(h : int) -> dict[str, dict]:
            dedup : dict[int, set[str]] = defaultdict(set)
            for k, h2 in self.cache[h].items():
                dedup[h2].add(k)

            return {"/".join(keys) : reconstruct_node(h) for h, keys in dedup.items()}
        return reconstruct_node(self.root_hash)
    
    def reconstruct_compressed_ecmwf_style(self, max_depth=None, from_node=None) -> dict[str, dict]:
        "Reconstruct the tree as a normal nested dictionary"
        def reconstruct_node(h : int, depth : int) -> dict[str, dict]:
            if max_depth is not None and depth > max_depth: 
                return {}
            dedup : dict[tuple[int, str], set[str]] = defaultdict(set)
            for k, h2 in self.cache[h].items():
                key, value = k.split("=")
                dedup[(h2, key)].add(value)

            return {f"{key}={','.join(values)}" : reconstruct_node(h, depth=depth+1) for (h, key), values in dedup.items()}
        return reconstruct_node(from_node or self.root_hash, depth=0)
    
    def to_json(self, max_depth=None, from_node=None) -> dict:
        def reconstruct_node(h : int, depth : int) -> list[JSONNode]:
            if max_depth is not None and depth > max_depth: 
                return {}
            dedup : dict[tuple[int, str], set[str]] = defaultdict(set)
            for k, h2 in self.cache[h].items():
                key, value = k.split("=")
                dedup[(h2, key)].add(value)

            return [JSONNode(
                key = key,
                values = list(values),
                children = reconstruct_node(h, depth=depth+1),
            ) for (h, key), values in dedup.items()]
        
        return asdict(reconstruct_node(from_node or self.root_hash, depth=0)[0])
    
    def __init__(self, tree : Tree):
        self.cache = {}
        self.empty_hash = hash(RefcountedDict({}))
        
        # Recursively cache the tree
        self.root_hash = self.cache_tree(tree)

        # Keep a reference to the root of the tree
        self.tree = self.cache[self.root_hash]


    def lookup(self, keys : tuple[str, ...]) -> tuple[bool, tuple[str, ...]]:
        """
        Lookup a subtree in the tree
        Returns success, path
        if success == True it means the path got to the bottom of the tree and path will be equal to keys
        if success == False, path will holds the keys that were found
        """
        loc = self.tree
        for i, key in enumerate(keys):
            if key in loc:
                h = loc[key] # get the hash of the subtree
                loc = self.cache[h] # get the subtree
            else:
                return False, keys[:i], h
        return True, keys, h

    def keys(self, keys : tuple[str, ...] = ()) -> list[str] | None:
        loc = self.tree
        for i, key in enumerate(keys):
            if key in loc:
                h = loc[key] # get the hash of the subtree
                loc = self.cache[h] # get the subtree
            else:
                return None
        return list(loc.keys())
    
    def multi_match(self, request : dict[str, list[str]], loc = None):
        if loc is None: loc = self.tree
        if loc == {}: return {"_END_" : {}}
        matches = {}
        for request_key, request_values in request.items():
            for request_value in request_values:
                meta_key = f"{request_key}={request_value}"
                if meta_key in loc:
                    new_loc = self.cache[loc[meta_key]]
                    matches[meta_key] = self.multi_match(request, new_loc)

        if not matches: return {k : {} for k in loc.items()}
        return matches


    def _insert(self, old_h : int, tree: RefcountedDict, keys : tuple[str, ...]) -> int:
        "Insert keys in the subtree and return the new hash of the subtree"
        key, *rest = keys
        assert old_h in self.cache

        # Adding a new branch to the tree
        if key not in tree:
            new_tree = RefcountedDict(tree | {key : self._cache_path(rest)})
        
        else:
            # Make a copy of the tree and update the subtree
            new_tree = RefcountedDict(tree.copy())
            subtree_h = tree[key]
            subtree = self.cache[subtree_h]
            new_tree[key] = self._insert(subtree_h, subtree, tuple(rest))
                
        # no-op if the hash hasn't changed
        new_h = self._replace_in_cache(old_h, new_tree)
        return new_h


    def insert(self, keys : tuple[str, ...]):
        """
        Insert a new branch into the compressed tree
        """
        already_there, path = self.lookup(keys)
        if already_there: 
            return
        # Update the tree
        self.root_hash = self._insert(self.root_hash, self.tree, keys)
        self.tree = self.cache[self.root_hash]

    def insert_tree(self, subtree: Tree):
        """
        Insert a whole tree into the compressed tree.
        """
        self.root_hash = self._insert_tree(self.root_hash, self.tree, subtree)
        self.tree = self.cache[self.root_hash]

    def _insert_tree(self, old_h: int, tree: RefcountedDict, subtree: Tree) -> int:
        """
        Recursively insert a subtree into the compressed tree and return the new hash.
        """
        assert old_h in self.cache

        # Make a copy of the tree to avoid modifying shared structures
        new_tree = RefcountedDict(tree.copy())
        for key, sub_subtree in subtree.items():
            if key not in tree:
                # Key is not in current tree, add the subtree
                # Cache the subtree rooted at sub_subtree
                subtree_h = self.cache_tree(sub_subtree)
                new_tree[key] = subtree_h
            else:
                # Key is in tree, need to recursively merge
                # Get the hash and subtree from the current tree
                child_h = tree[key]
                child_tree = self.cache[child_h]
                # Recursively merge
                new_child_h = self._insert_tree(child_h, child_tree, sub_subtree)
                new_tree[key] = new_child_h

        # Replace the old hash with the new one in the cache
        new_h = self._replace_in_cache(old_h, new_tree)
        return new_h

    def save(self, path : Path):
        "Save the compressed tree to a file"
        with open(path, "w") as f:
            json.dump({
                "cache" : {k : {"refcount" : v.refcount, "dict" : v} for k, v in self.cache.items()},
                "root_hash": self.root_hash
            }, f)

    @classmethod
    def load(cls, path : Path) -> "CompressedTree":
        "Load the compressed tree from a file"
        with open(path) as f:
            data = json.load(f)
        return cls.from_json(data)


    @classmethod
    def from_json(cls, data : dict) -> "CompressedTree":
        c = CompressedTree({})
        c.cache = {}
        for k, v in data["cache"].items():
            c.cache[int(k)] = RefcountedDict(v["dict"])
            c.cache[int(k)].refcount = v["refcount"]

        c.root_hash = data["root_hash"]
        c.tree = c.cache[c.root_hash]
        return c


if __name__ == "__main__":
    original_tree = {
        "a": {
            "b1": {
                "c": {}
            },
            "b2" : {
                "c": {}
            },
            "b3*": {
                "c*": {}
            }
        }
    }
    
    c_tree = CompressedTree(original_tree)

    assert c_tree.lookup(("a", "b1", "c")) == (True, ("a", "b1", "c"))
    assert c_tree.lookup(("a", "b1", "d")) == (False, ("a", "b1"))

    print(json.dumps(c_tree.reconstruct_compressed(), indent = 4))

    assert c_tree.reconstruct() == original_tree

    c_tree.insert(("a", "b1", "d"))
    c_tree.insert(("a", "b2", "d"))
    print(json.dumps(c_tree.reconstruct(), indent = 4))

    print(json.dumps(c_tree.reconstruct_compressed(), indent = 4))
    print(c_tree.cache)

    # test round trip
    assert CompressedTree(original_tree).reconstruct() == original_tree

    # test adding a key 
    added_keys_tree = {
        "a": {
            "b1": {
                "c": {}
            },
            "b2" : {
                "c": {},
                "d" : {}
            },
            "b3*": {
                "c*": {},
                "d*": {}
            }
        }
    }
    c_tree = CompressedTree(original_tree)
    c_tree.insert(("a", "b2", "d"))
    c_tree.insert(("a", "b3*", "d*"))
    assert c_tree.reconstruct() == added_keys_tree

    print(c_tree.reconstruct_compressed())


import dataclasses
import json
from dataclasses import dataclass, field
from typing import Any

import pe
from pe.actions import Pack
from pe.operators import Class, Star

from .fdb_types import FDB_type_to_implementation, FDBType


@dataclass(frozen=True)
class KeySpec:
    """
    Represents the specification of a single key in an FDB schema file. For example in
    ```
    [ class, expver, stream=lwda, date, time, domain?
       [ type=ofb/mfb/oai
               [ obsgroup, reportype ]]]
    ```
    class, expver, type=ofdb/mfb/oai etc are the KeySpecs

    These can have additional information such as: flags like `domain?`, allowed values like `type=ofb/mfb/oai`
    or specify type information with `date: ClimateMonthly`

    """

    key: str
    type: FDBType = field(default_factory=FDBType)
    flag: str | None = None
    values: tuple = field(default_factory=tuple)
    comment: str = ""

    def __repr__(self):
        repr = self.key
        if self.flag:
            repr += self.flag
        # if self.type:
        #     repr += f":{self.type}"
        if self.values:
            repr += "=" + "/".join(self.values)
        return repr

    def matches(self, key, value):
        # Sanity check!
        if self.key != key:
            return False

        # Some keys have a set of allowed values type=ofb/mfb/oai
        if self.values:
            if value not in self.values:
                return False

        # Check the formatting of values like Time or Date
        if self.type and not self.type.validate(value):
            return False

        return True

    def is_optional(self):
        if self.flag is None:
            return False
        return "?" in self.flag

    def is_allable(self):
        if self.flag is None:
            return False
        return "*" in self.flag


@dataclass(frozen=True)
class Comment:
    "Represents a comment node in the schema"

    value: str


@dataclass(frozen=True)
class FDBSchemaTypeDef:
    "Mapping between FDB schema key names and FDB Schema Types, i.e expver is of type Expver"

    key: str
    type: str


# This is the schema grammar written in PEG format
fdb_schema = pe.compile(
    r"""
    FDB < Line+ EOF
    Line < Schema / Comment / TypeDef / empty

    # Comments
    Comment <- "#" ~non_eol*
    non_eol              <- [\x09\x20-\x7F] / non_ascii
    non_ascii            <- [\x80-\uD7FF\uE000-\U0010FFFF]

    # Default Type Definitions
    TypeDef < String ":" String ";"

    # Schemas are the main attraction
    # They're a tree of KeySpecs.
    Schema < "[" KeySpecs (","? Schema)* "]"

    # KeySpecs can be just a name i.e expver
    # Can also have a type expver:int
    # Or a flag expver?
    # Or values expver=xxx
    KeySpecs < KeySpec_ws ("," KeySpec_ws)*
    KeySpec_ws < KeySpec
    KeySpec <- key:String (flag:Flag)? (type:Type)? (values:Values)? ([ ]* comment:Comment)?
    Flag <- ~("?" / "-" / "*")
    Type <- ":" [ ]* String
    Values <- "=" Value ("/" Value)*

    # Low level stuff 
    Value   <- ~([-a-zA-Z0-9_]+)
    String   <- ~([a-zA-Z0-9_]+)
    EOF  <- !.
    empty <- ""
    """,
    actions={
        "Schema": Pack(tuple),
        "KeySpec": KeySpec,
        "Values": Pack(tuple),
        "Comment": Comment,
        "TypeDef": FDBSchemaTypeDef,
    },
    ignore=Star(Class("\t\f\r\n ")),
    # flags=pe.DEBUG,
)


def post_process(entries):
    "Take the raw output from the PEG parser and split it into type definitions and schema entries."
    typedefs = {}
    schemas = []
    for entry in entries:
        match entry:
            case c if isinstance(c, Comment):
                pass
            case t if isinstance(t, FDBSchemaTypeDef):
                typedefs[t.key] = t.type
            case s if isinstance(s, tuple):
                schemas.append(s)
            case _:
                raise ValueError
    return typedefs, tuple(schemas)


def determine_types(types, node):
    "Recursively walk a schema tree and insert the type information."
    if isinstance(node, tuple):
        return [determine_types(types, n) for n in node]
    return dataclasses.replace(node, type=types.get(node.key, FDBType()))


@dataclass
class Key:
    key: str
    value: Any
    key_spec: KeySpec
    reason: str

    def str_value(self):
        return self.key_spec.type.format(self.value)

    def __bool__(self):
        return self.reason in {"Matches", "Skipped", "Select All"}

    def emoji(self):
        return {"Matches": "✅", "Skipped": "⏭️", "Select All": "★"}.get(
            self.reason, "❌"
        )

    def info(self):
        return f"{self.emoji()} {self.key:<12}= {str(self.value):<12} ({self.key_spec}) {self.reason if not self else ''}"

    def __repr__(self):
        return f"{self.key}={self.key_spec.type.format(self.value)}"

    def as_json(self):
        return dict(
            key=self.key,
            value=self.str_value(),
            reason=self.reason,
        )


class FDBSchema:
    """
    Represents a parsed FDB Schema file.
    Has methods to validate and convert request dictionaries to a mars request form with validation and type information.
    """

    def __init__(self, string, defaults: dict[str, str] = {}):
        """
        1. Use a PEG parser on a schema string,
        2. Separate the output into schemas and typedefs
        3. Insert any concrete implementations of types from fdb_types.py defaulting to generic string type
        4. Walk the schema tree and annotate it with type information.
        """
        m = fdb_schema.match(string)
        g = list(m.groups())
        self._str_types, schemas = post_process(g)
        self.types = {
            key: FDB_type_to_implementation[type]
            for key, type in self._str_types.items()
        }
        self.schemas = determine_types(self.types, schemas)
        self.defaults = defaults

    def __repr__(self):
        return json.dumps(
            dict(schemas=self.schemas, defaults=self.defaults), indent=4, default=repr
        )

    @classmethod
    def consume_key(
        cls, key_spec: KeySpec, request: dict[str, Any]
    ) -> Key:
        key = key_spec.key
        try:
            value = request[key]
        except KeyError:
            if key_spec.is_optional():
                return Key(key_spec.key, "", key_spec, "Skipped")
            if key_spec.is_allable():
                return Key(key_spec.key, "", key_spec, "Select All")
            else:
                return Key(
                    key_spec.key, "", key_spec, "Key Missing"
                )

        if key_spec.matches(key, value):
            return Key(
                key_spec.key,
                key_spec.type.parse(value),
                key_spec,
                "Matches",
            )
        else:
            return Key(
                key_spec.key, value, key_spec, "Incorrect Value"
            )

    @classmethod
    def _DFS_match(
        cls, tree: list, request: dict[str, Any]
    ) -> tuple[bool | list, list[Key]]:
        """Do a DFS on the schema tree, returning the deepest matching path
        At each stage return whether we matched on this path, and the path itself.

        When traversing the tree there are three cases to consider:
        1. base case []
        2. one schema [k, k, k, [k, k, k]]
        3. list of schemas [[k,k,k], [k,k,k], [k,k,k]]
        """
        #  Case 1: Base Case
        if not tree:
            return True, []

        # Case 2: [k, k, k, [k, k, k]]
        if isinstance(tree[0], KeySpec):
            node, *tree = tree
            # Check if this node is in the request
            match_result = cls.consume_key(node, request)

            # If if isn't then terminate this path here
            if not match_result:
                return False, [match_result,]  # fmt: skip

            # Otherwise continue walking the tree and return the best result
            matched, path = cls._DFS_match(tree, request)

            # Don't put the key in the path if it's optional and we're skipping it.
            if match_result.reason != "Skipped":
                path = [match_result,] + path  # fmt: skip

            return matched, path

        # Case 3: [[k, k, k], [k, k, k]]
        branches = []
        for branch in tree:
            matched, branch_path = cls._DFS_match(branch, request)

            # If this branch matches, terminate the DFS and use this.
            if matched:
                return branch, branch_path
            else:
                branches.append(branch_path)

        # If no branch matches, return the one with the deepest match
        return False, max(branches, key=len)

    @classmethod
    def _DFS_match_all(
        cls, tree: list, request: dict[str, Any]
    ) -> list[list[Key]]:
        """Do a DFS on the schema tree, returning all matching paths or partial matches.
        At each stage return all matching paths and the deepest partial matches.

        When traversing the tree there are three cases to consider:
        1. base case []
        2. one schema [k, k, k, [k, k, k]]
        3. list of schemas [[k,k,k], [k,k,k], [k,k,k]]
        """
        # Case 1: Base Case
        if not tree:
            return [[]]

        # Case 2: [k, k, k, [k, k, k]]
        if isinstance(tree[0], KeySpec):
            node, *tree = tree
            # Check if this node is in the request
            request_values = request.get(node.key, None)

            if request_values is None:
                # If the key is not in the request, return a partial match with Key Missing
                return [[Key(node.key, "", node, "Key Missing")]]

            # If the request value is a list, try to match each value
            if isinstance(request_values, list):
                all_matches = []
                for value in request_values:
                    match_result = cls.consume_key(node, {node.key: value})

                    if match_result:
                        sub_matches = cls._DFS_match_all(tree, request)
                        for match in sub_matches:
                            if match_result.reason != "Skipped":
                                match.insert(0, match_result)
                            all_matches.append(match)

                return all_matches if all_matches else [[Key(node.key, "", node, "No Match Found")]]
            else:
                # Handle a single value
                match_result = cls.consume_key(node, request)

                # If it isn't then return a partial match with Key Missing
                if not match_result:
                    return [[Key(node.key, "", node, "Key Missing")]]

                # Continue walking the tree and get all matches
                all_matches = cls._DFS_match_all(tree, request)

                # Prepend the current match to all further matches
                for match in all_matches:
                    if match_result.reason != "Skipped":
                        match.insert(0, match_result)

                return all_matches

        # Case 3: [[k, k, k], [k, k, k]]
        all_branch_matches = []
        for branch in tree:
            branch_matches = cls._DFS_match_all(branch, request)
            all_branch_matches.extend(branch_matches)

        # Return all of the deepest partial matches or complete matches
        return all_branch_matches

    def match_all(self, request: dict[str, Any]):
        request = request | self.defaults
        return self._DFS_match_all(self.schemas, request)

    def match(self, request: dict[str, Any]):
        request = request | self.defaults
        return self._DFS_match(self.schemas, request)


class FDBSchemaFile(FDBSchema):
    def __init__(self, path: str):
        with open(path, "r") as f:
            return super().__init__(f.read())



import dataclasses
from collections import defaultdict
from dataclasses import dataclass, field

from frozendict import frozendict

from .Qube import Enum, NodeData, Tree
from .tree_formatters import HTML, node_tree_to_html, node_tree_to_string

NodeId = int
CacheType = dict[NodeId, "CompressedNode"]

@dataclass(frozen=True)
class CompressedNode:
    id: NodeId = field(hash=False, compare=False)
    data: NodeData

    _children: tuple[NodeId, ...]
    _cache: CacheType = field(repr=False, hash=False, compare=False)

    @property
    def children(self) -> tuple["CompressedNode", ...]:
        return tuple(self._cache[i] for i in self._children)

    def summary(self, debug = False) -> str:
        if debug:  return f"{self.data.key}={self.data.values.summary()} ({self.id})"
        return f"{self.data.key}={self.data.values.summary()}" if self.data.key != "root" else "root"


@dataclass(frozen=True)
class CompressedTree:
    """
    This tree is compressed in two distinct different ways:
    1. Product Compression: Nodes have a key and **multiple values**, so each node represents many logical nodes key=value1, key=value2, ...
       Each of these logical nodes is has identical children so we can compress them like this.
       In this way any distinct path through the tree represents a cartesian product of the values, otherwise known as a datacube.

    2. In order to facilitate the product compression described above we need to know when two nodes have identical children.
       To do this every node is assigned an Id which is initially computed as a hash from the nodes data and its childrens' ids.
       In order to avoid hash collisions we increment the initial hash if it's already in the cache for a different node 
       we do this until we find a unique id.

       Crucially this allows us to later determine if a new node is already cached: 
        id = hash(node)
        while True:
            if id not in cache: The node is definitely not in the cache
            elif cache[id] != node: Hash collision, increment id and try again
            else: The node is already in the cache
            id += 1

    This tree can be walked from the root by repeatedly looking up the children of a node in the cache.

    This structure facilitates compression because we can look at the children of a node:
        If two chidren have the same key, metadata and children then we can compress them into a single node.

"""
    root: CompressedNode
    cache: CacheType

    @staticmethod
    def add_to_cache(cache : dict[NodeId, CompressedNode], data : NodeData, _children: tuple[NodeId, ...]) -> NodeId:
        """
        This function is responsible for adding a new node to the cache and returning its id.
        Crucially we need a way to check if new nodes are already in the cache, so we hash them.
        But in case of a hash collision we need to increment the id and try again.
        This way we will always eventually find a unique id for the node.
        And we will never store the same node twice with a different id.
        """
        _children = tuple(sorted(_children))
        id = hash((data, _children))

        # To avoid hash collisions, we increment the id until we find a unique one
        tries = 0
        while True:
            tries += 1
            if id not in cache:
                # The node isn't in the cache and this id is free
                cache[id] = CompressedNode(id = id,
                                           data = data,
                                           _children = _children,
                                           _cache = cache)
                break 
            
            if cache[id].data == data and cache[id]._children == _children:
                break # The node is already in the cache

            # This id is already in use by a different node so increment it (mod) and try again
            id = (id + 1) % (2**64)

            if tries > 100:
                raise RuntimeError("Too many hash collisions, something is wrong.")
        
        return id


    @classmethod
    def from_tree(cls, tree : Tree) -> 'CompressedTree':
        cache = {}

        def cache_tree(level : Tree) -> NodeId:
            node_data = NodeData(
                key = level.key,
                values = level.values,
            )

            # Recursively cache the children
            children = tuple(cache_tree(c) for c in level.children)
            
            # Add the node to the cache and return its id
            return cls.add_to_cache(cache, node_data, children)
        
        root = cache_tree(tree)
        return cls(cache = cache, root = cache[root])
    
    def __str__(self, depth=None) -> str:
        return "".join(node_tree_to_string(self.root, depth = depth))
    
    def print(self, depth = None): print(self.__str__(depth = depth))
    
    def html(self, depth = 2, debug = False) -> HTML:
        return HTML(node_tree_to_html(self.root, depth = depth, debug = debug))
    
    def _repr_html_(self) -> str:
        return node_tree_to_html(self.root, depth = 2)
    
    def __getitem__(self, args) -> 'CompressedTree':
        key, value = args
        for c in self.root.children:
            if c.data.key == key and value in c.data.values:
                data = dataclasses.replace(c.data, values = Enum((value,)))
                return CompressedTree(
                    cache = self.cache,
                    root = dataclasses.replace(c, data = data)
                )
        raise KeyError(f"Key {key} not found in children.")
    
    def collapse_children(self, node: "CompressedNode") -> "CompressedNode":
        # First perform the collapse on the children
        new_children = [self.collapse_children(child) for child in node.children]

        # Now take the set of new children and see if any have identical key, metadata and children
        # the values may different and will be collapsed into a single node
        identical_children = defaultdict(set)
        for child in new_children:
            identical_children[(child.data.key, child.data.metadata, child._children)].add(child)
        
        # Now go through and create new compressed nodes for any groups that need collapsing
        new_children = []
        for (key, metadata, _children), child_set in identical_children.items():
            if len(child_set) > 1:
                # Compress the children into a single node
                assert all(isinstance(child.data.values, Enum) for child in child_set), "All children must have Enum values"
                node_data = NodeData(
                    key = key,
                    metadata = frozendict(), # Todo: Implement metadata compression
                    values = Enum(tuple(v for child in child_set for v in child.data.values.values)),
                )
                
                # Add the node to the cache
                id = type(self).add_to_cache(self.cache, node_data, _children)
            else:
                # If the group is size one just keep it
                id = child_set.pop().id
            
            new_children.append(id)

        id = self.add_to_cache(self.cache, node.data, tuple(sorted(new_children)))
        return self.cache[id]


    def compress(self) -> 'CompressedTree':
        return CompressedTree(cache = self.cache, root = self.collapse_children(self.root))

    def lookup(self, selection : dict[str, str]):
        nodes = [self.root]
        for _ in range(1000):
            found = False
            current_node = nodes[-1]
            for c in current_node.children:
                if selection.get(c.data.key, None) in c.data.values:
                    if found: 
                        raise RuntimeError("This tree is invalid, because it contains overlapping branches.")
                    nodes.append(c)
                    selection.pop(c.data.key)
                    found = True
            
            if not found:
                return nodes
            
        raise RuntimeError("Maximum node searches exceeded, the tree contains a loop or something is buggy.")



    
    # def reconstruct(self) -> Tree:
    #     def reconstruct_node(h : int) -> Tree:
    #         node = self.cache[h]
    #         dedup : dict[tuple[int, str], set[NodeId]] = defaultdict(set)
    #         for index in self.cache[h].children:
    #             child_node = self.cache[index]
    #             child_hash = hash(child_node.children)
    #             assert isinstance(child_node.values, Enum)
    #             dedup[(child_hash, child_node.key)].add(index)

        
    #         children = tuple(
    #             Tree(key = key, values = Enum(tuple(values)), 
    #             children = tuple(reconstruct_node(i) for i in self.cache[next(indices)].children)
    #             )
    #             for (_, key), indices in dedup.items()
    #         )

    #         return Tree(
    #             key = node.key,
    #             values = node.values,
    #             children = children,
    #         )
    #     return reconstruct_node(self.root)


#! .venv/bin/python

import redis
import yaml
import json

print("Opening redis connection")
r = redis.Redis(host="redis", port=6379, db=0)

print("Loading data from local files")
with open("config/climate-dt/compressed_tree.json") as f:
    compressed_catalog = json.load(f)

with open("config/climate-dt/language.yaml") as f:
    mars_language = yaml.safe_load(f)["_field"]

print("Storing data in redis")
r.set('compressed_catalog', json.dumps(compressed_catalog))
r.set('mars_language', json.dumps(mars_language))